import matplotlib.pyplot as plt
import numpy as np


def quasi_Newton(x0,f,eps,h):
    # Version 1d
    itmax=20
    it=0
    x=x0
    fx=f(x0) 
    liste_x=np.zeros(itmax+1)
    liste_fx=np.zeros(itmax+1)
    liste_x[0]=x0
    liste_fx[0]=f(x0) 
    while np.abs(fx) > eps and it < itmax:
        it+=1
        dfx=(f(x+h)-f(x))/h
        x=x-f(x)/dfx
        fx=f(x)
        #crit=np.abs(f(x))
        liste_x[it]=x
        liste_fx[it]=fx
    return liste_x,liste_fx,it

def quasi_Newton_vec(x0,f,df,eps,h):
    # Version vectorielle
    itmax=20
    it=0
    x=x0
    crit=np.linalg.norm(f(x0))
    liste_it=np.zeros(itmax+1)
    liste_crit=np.zeros(itmax+1)
    liste_x=np.zeros(itmax+1)
    liste_y=np.zeros(itmax+1)
    liste_it[0]=0
    liste_crit[0]=crit
    liste_x[0]=x[0]
    liste_y[0]=x[1]
    while crit > eps and it < itmax:
        it+=1
        x=x-np.linalg.solve(df(x,h),f(x))
        crit=np.linalg.norm(f(x))
        liste_it[it]=it
        liste_crit[it]=crit
        liste_x[it]=x[0]
        liste_y[it]=x[1]
    return liste_x,liste_y,liste_crit,liste_it,it



def f2(x):
    x1=x[0]
    x2=x[1]
    x1=x1**2
    x2=x2**2
    y1=x1+x2-2
    y2=x1-x2-1
    return y1,y2

def df2(x,h):
    x1=x[0]**2
    xx1=(x[0]+h)**2
    x2=x[1]**2
    xx2=(x[1]+h)**2
    y11=(xx1-x1)/h
    y12=(xx2-x2)/h
    y21=(xx1-x1)/h
    y22=(-xx2+x2)/h
    y1=[y11,y12]
    y2=[y21,y22]
    return y1,y2






def f1(x):
    return np.exp(x)-np.e

def df1(x):
    return np.exp(x)

# Test de quasi-Newton avec f1

x0=-1
f=f1
df=df1
eps=1e-6
h=0.1

tabx,tabfx,it=quasi_Newton(x0,f,eps,h) 
print('La méthode converge en ',it,'pas')

tabxeff=tabx[:it]
tabfxeff=tabfx[:it]

t=np.linspace(-2,5,100) # Graphe de f1
yt=f1(t)
plt.plot(t,yt,tabxeff,tabfxeff,tabxeff[-1],tabfxeff[-1],'*')
plt.title('Itérés quasi-Newton pour f1, x0=-1, h=0.1')
plt.show()


h=0.001
tabx,tabfx,it=quasi_Newton(x0,f,eps,h) # Newton4(x0,f,df,eps)
tabxeff=tabx[:it]
tabfxeff=tabfx[:it]
print('La méthode converge en ',it,'pas')
plt.plot(t,yt,tabxeff,tabfxeff,tabxeff[0],tabfxeff[0],'o',tabxeff[-1],tabfxeff[-1],'*')
plt.title('Itérés quasi-Newton pour f1, x0=-1, h=0.01')
plt.show()


tabit=np.zeros((2,5))
tabi=np.arange(1,6)
i=0
for p in range(1,6):
    h=10**(-p)
    tabx,tabfx,it=quasi_Newton(x0,f,eps,h)
    tabit[0][i]=p
    tabit[1][i]=it
    i+=1
    
print('Evolution du nombre des itérations en fonction de p tel que h=10**(-p):',tabit)    


h=0.001

tabx,tabfx,it=quasi_Newton(x0,f,eps,h) 
print('La méthode converge en ',it,'pas')

tabxeff=tabx[:it]
tabfxeff=tabfx[:it]
err=tabxeff-tabxeff[-1]
errx=err[1:-1]
erry=err[2:]
plt.plot(errx,erry,errx[-1],erry[-1],'*')
plt.title('Croissance quadratique de l Erreur quasi-Newton pour f1, x0=-1,h=0.001')
plt.show()


errx2=errx
lx=len(errx)
for i in range(lx):
    errx2[i]=errx[i]**2
    
plt.plot(errx2,erry,errx2[-1],erry[-1],'*')
plt.title('Comparaison avec le carré  des erreurs: on cherche une droite')
plt.show()
    
lerrx=np.log(errx[:-1])
lerry=np.log(erry[:-1])

plt.plot(lerrx,lerry,lerrx[-1],lerry[-1],'*')
plt.title('Log erreur pour f1, x0=-1,h=0.001')
plt.show()


A=np.polyfit(lerrx,lerry,1)
print('Pente de log erreur pour f1, x0=-1,h=0.001: ',A[0])


####### Quasi Newton vectoriel

f=f2
df=df2
eps=1e-6
x0=np.array([-3,5])

print('Pour f2 et x0=[-3,5]:')

for p in range(1,6):
    h=10**(-p)
    tabx,taby,tabcrit,tabit,it=quasi_Newton_vec(x0,f,df,eps,h)
    print('La méthode converge en ',it,'pas')




#Graphe de l'intersection entre le cercle et l'hyperbole 

t=np.linspace(0,2*np.pi,100) # le cercle
r=np.sqrt(2)
xt=r*np.cos(t)
yt=r*np.sin(t)


s=np.linspace(-2,2,600) # l'hyperbole
xs=np.cosh(s)
ys=np.sinh(s)


h=0.001
x0=np.array([-3,5])

tabx,taby,tabcrit,tabit,it=quasi_Newton_vec(x0,f,df,eps,h)
print('La méthode converge en ',it,'pas')


tabcriteff=tabcrit[:it]
tabiteff=tabit[:it]

plt.plot(tabiteff,tabcriteff,tabiteff[-1],tabcriteff[-1],'*')
plt.title('Evolution du critere d arrêt pour f2 et x0=[-3,5], h=0.001')
plt.show()

tabxeff=tabx[:it]
tabyeff=taby[:it]
plt.plot(xt,yt,xs,ys,-xs,ys,tabxeff,tabyeff)
plt.title('Itérés pour f2 et x0=[-3,5], h=0.001')
plt.show()

err=np.zeros(it)
for i in range(it):
    err[i]=np.linalg.norm(np.array([tabxeff[i]-tabxeff[-1],tabyeff[i]-tabyeff[-1]]))
errx=err[1:-1]
erry=err[2:]
plt.plot(errx,erry,errx[-1],erry[-1],'*')
plt.title('Croissance quadratique de l erreur pour f2 et x0=[-3,5], h=0.001')
plt.show()

errx2=errx
lx=len(errx)
for i in range(lx):
    errx2[i]=errx[i]**2
    
plt.plot(errx2,erry,errx2[-1],erry[-1],'*')
plt.title('Comparaison avec le carré de l erreur  pour f2 et x0=[-3,5], h=0.001 ')
plt.show()
    
lerrx=np.log(errx[:-1])
lerry=np.log(erry[:-1])

plt.plot(lerrx,lerry,lerrx[-1],lerry[-1],'*')
plt.title('Log-erreur pour f2 et x0=[-3,5], h=0.001')
plt.show()

A=np.polyfit(lerrx,lerry,1)
print('Pente de log erreur pour f2 et x0=[-3,5], h=0.001:',A[0])




    


